@@ -4,7 +4,6 @@
 
 from __future__ import with_statement
 
-import collections
 import errno
 import filecmp
 import os.path
@@ -12,6 +11,8 @@ import re
 import tempfile
 import sys
 
+from six.moves import collections_abc
+
 
 # A minimal memoizing decorator. It'll blow up if the args aren't immutable,
 # among other "problems".
@@ -494,7 +495,7 @@ def uniquer(seq, idfun=None):
 
 
 # Based on http://code.activestate.com/recipes/576694/.
-class OrderedSet(collections.MutableSet):
+class OrderedSet(collections_abc.MutableSet):
   def __init__(self, iterable=None):
     self.end = end = []
     end += [None, end, end]         # sentinel node for doubly linked list
